# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from multiprocessing import Manager
from typing import List

from swh.model.model import BaseModel

from .kafka import ModelObject

logger = logging.getLogger(__name__)


class InMemoryJournalWriter:
    def __init__(self):
        # Share the list of objects across processes, for RemoteAPI tests.
        self.manager = Manager()
        self.objects = self.manager.list()
        self.privileged_objects = self.manager.list()

    def write_addition(
        self, object_type: str, object_: ModelObject, privileged: bool = False
    ) -> None:
        assert isinstance(object_, BaseModel)
        if privileged:
            self.privileged_objects.append((object_type, object_))
        else:
            self.objects.append((object_type, object_))

    write_update = write_addition

    def write_additions(
        self, object_type: str, objects: List[ModelObject], privileged: bool = False
    ) -> None:
        for object_ in objects:
            self.write_addition(object_type, object_, privileged)



# Copyright (C) 2016-2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

import click

from swh.core.cli import CONTEXT_SETTINGS


@click.group(name="journal", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(exists=True, dir_okay=False,),
    help="Configuration file.",
)
@click.pass_context
def cli(ctx, config_file):
    """DEPRECATED Software Heritage Journal tools.
    """
    pass


@cli.command()
@click.option(
    "--stop-after-objects",
    "-n",
    default=None,
    type=int,
    help="Stop after processing this many objects. Default is to " "run forever.",
)
@click.pass_context
def replay(ctx, stop_after_objects):
    """DEPRECATED: use `swh storage replay` instead.

    Requires swh.storage >= 0.0.188.
    """
    ctx.fail("DEPRECATED")


@cli.command()
@click.argument("object_type")
@click.option("--start-object", default=None)
@click.option("--end-object", default=None)
@click.option("--dry-run", is_flag=True, default=False)
@click.pass_context
def backfiller(ctx, object_type, start_object, end_object, dry_run):
    """DEPRECATED: use `swh storage backfill` instead.

    Requires swh.storage >= 0.0.188.

    """
    ctx.fail("DEPRECATED")


@cli.command("content-replay")
@click.option(
    "--stop-after-objects",
    "-n",
    default=None,
    type=int,
    help="Stop after processing this many objects. Default is to " "run forever.",
)
@click.option(
    "--exclude-sha1-file",
    default=None,
    type=click.File("rb"),
    help="File containing a sorted array of hashes to be excluded.",
)
@click.option(
    "--check-dst/--no-check-dst",
    default=True,
    help="Check whether the destination contains the object before " "copying.",
)
@click.pass_context
def content_replay(ctx, stop_after_objects, exclude_sha1_file, check_dst):
    """DEPRECATED: use `swh objstorage replay` instead.

    This needs the swh.objstorage.replayer package."""
    ctx.fail("DEPRECATED")


def main():
    logging.basicConfig()
    return cli(auto_envvar_prefix="SWH_JOURNAL")


if __name__ == "__main__":
    main()



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import warnings


def get_journal_writer(cls, **kwargs):
    if "args" in kwargs:
        warnings.warn(
            'Explicit "args" key is deprecated, use keys directly instead.',
            DeprecationWarning,
        )
        kwargs = kwargs["args"]

    if cls == "inmemory":  # FIXME: Remove inmemory in due time
        warnings.warn(
            "cls = 'inmemory' is deprecated, use 'memory' instead", DeprecationWarning
        )
        cls = "memory"
    if cls == "memory":
        from .inmemory import InMemoryJournalWriter as JournalWriter
    elif cls == "kafka":
        from .kafka import KafkaJournalWriter as JournalWriter
    else:
        raise ValueError("Unknown journal writer class `%s`" % cls)

    return JournalWriter(**kwargs)



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import time
from typing import Dict, Iterable, List, NamedTuple, Optional, Type

from confluent_kafka import Producer, KafkaException

from swh.journal.serializers import (
    KeyType,
    ModelObject,
    object_key,
    pprint_key,
    key_to_kafka,
    value_to_kafka,
)

logger = logging.getLogger(__name__)


class DeliveryTag(NamedTuple):
    """Unique tag allowing us to check for a message delivery"""

    topic: str
    kafka_key: bytes


class DeliveryFailureInfo(NamedTuple):
    """Verbose information for failed deliveries"""

    object_type: str
    key: KeyType
    message: str
    code: str


def get_object_type(topic: str) -> str:
    """Get the object type from a topic string"""
    return topic.rsplit(".", 1)[-1]


class KafkaDeliveryError(Exception):
    """Delivery failed on some kafka messages."""

    def __init__(self, message: str, delivery_failures: Iterable[DeliveryFailureInfo]):
        self.message = message
        self.delivery_failures = list(delivery_failures)

    def pretty_failures(self) -> str:
        return ", ".join(
            f"{f.object_type} {pprint_key(f.key)} ({f.message})"
            for f in self.delivery_failures
        )

    def __str__(self):
        return f"KafkaDeliveryError({self.message}, [{self.pretty_failures()}])"


class KafkaJournalWriter:
    """This class is used to write serialized versions of swh.model.model objects to a
    series of Kafka topics.

    Topics used to send objects representations are built from a ``prefix`` plus the
    type of the object:

      ``{prefix}.{object_type}``

    Objects can be sent as is, or can be anonymized. The anonymization feature, when
    activated, will write anonymized versions of model objects in the main topic, and
    stock (non-anonymized) objects will be sent to a dedicated (privileged) set of
    topics:

      ``{prefix}_privileged.{object_type}``

    The anonymization of a swh.model object is the result of calling its
    ``BaseModel.anonymize()`` method. An object is considered anonymizable if this
    method returns a (non-None) value.

    Args:
      brokers: list of broker addresses and ports.
      prefix: the prefix used to build the topic names for objects.
      client_id: the id of the writer sent to kafka.
      producer_config: extra configuration keys passed to the `Producer`.
      flush_timeout: timeout, in seconds, after which the `flush` operation
        will fail if some message deliveries are still pending.
      producer_class: override for the kafka producer class.
      anonymize: if True, activate the anonymization feature.

    """

    def __init__(
        self,
        brokers: Iterable[str],
        prefix: str,
        client_id: str,
        producer_config: Optional[Dict] = None,
        flush_timeout: float = 120,
        producer_class: Type[Producer] = Producer,
        anonymize: bool = False,
    ):
        self._prefix = prefix
        self._prefix_privileged = f"{self._prefix}_privileged"
        self.anonymize = anonymize

        if not producer_config:
            producer_config = {}

        if "message.max.bytes" not in producer_config:
            producer_config = {
                "message.max.bytes": 100 * 1024 * 1024,
                **producer_config,
            }

        self.producer = producer_class(
            {
                "bootstrap.servers": ",".join(brokers),
                "client.id": client_id,
                "on_delivery": self._on_delivery,
                "error_cb": self._error_cb,
                "logger": logger,
                "acks": "all",
                **producer_config,
            }
        )

        # Delivery management
        self.flush_timeout = flush_timeout

        # delivery tag -> original object "key" mapping
        self.deliveries_pending: Dict[DeliveryTag, KeyType] = {}

        # List of (object_type, key, error_msg, error_name) for failed deliveries
        self.delivery_failures: List[DeliveryFailureInfo] = []

    def _error_cb(self, error):
        if error.fatal():
            raise KafkaException(error)
        logger.info("Received non-fatal kafka error: %s", error)

    def _on_delivery(self, error, message):
        (topic, key) = delivery_tag = DeliveryTag(message.topic(), message.key())
        sent_key = self.deliveries_pending.pop(delivery_tag, None)

        if error is not None:
            self.delivery_failures.append(
                DeliveryFailureInfo(
                    get_object_type(topic), sent_key, error.str(), error.name()
                )
            )

    def send(self, topic: str, key: KeyType, value):
        kafka_key = key_to_kafka(key)
        self.producer.produce(
            topic=topic, key=kafka_key, value=value_to_kafka(value),
        )

        self.deliveries_pending[DeliveryTag(topic, kafka_key)] = key

    def delivery_error(self, message) -> KafkaDeliveryError:
        """Get all failed deliveries, and clear them"""
        ret = self.delivery_failures
        self.delivery_failures = []

        while self.deliveries_pending:
            delivery_tag, orig_key = self.deliveries_pending.popitem()
            (topic, kafka_key) = delivery_tag
            ret.append(
                DeliveryFailureInfo(
                    get_object_type(topic),
                    orig_key,
                    "No delivery before flush() timeout",
                    "SWH_FLUSH_TIMEOUT",
                )
            )

        return KafkaDeliveryError(message, ret)

    def flush(self):
        start = time.monotonic()

        self.producer.flush(self.flush_timeout)

        while self.deliveries_pending:
            if time.monotonic() - start > self.flush_timeout:
                break
            self.producer.poll(0.1)

        if self.deliveries_pending:
            # Delivery timeout
            raise self.delivery_error(
                "flush() exceeded timeout (%ss)" % self.flush_timeout,
            )
        elif self.delivery_failures:
            raise self.delivery_error("Failed deliveries after flush()")

    def _sanitize_object(
        self, object_type: str, object_: ModelObject
    ) -> Dict[str, str]:
        dict_ = object_.to_dict()
        if object_type == "origin_visit":
            # :(
            dict_["date"] = str(dict_["date"])
        if object_type == "content":
            dict_.pop("data", None)
        return dict_

    def _write_addition(self, object_type: str, object_: ModelObject) -> None:
        """Write a single object to the journal"""
        key = object_key(object_type, object_)

        if self.anonymize:
            anon_object_ = object_.anonymize()
            if anon_object_:  # can be either None, or an anonymized object
                # if the object is anonymizable, send the non-anonymized version in the
                # privileged channel
                topic = f"{self._prefix_privileged}.{object_type}"
                dict_ = self._sanitize_object(object_type, object_)
                logger.debug("topic: %s, key: %s, value: %s", topic, key, dict_)
                self.send(topic, key=key, value=dict_)
                object_ = anon_object_

        topic = f"{self._prefix}.{object_type}"
        dict_ = self._sanitize_object(object_type, object_)
        logger.debug("topic: %s, key: %s, value: %s", topic, key, dict_)
        self.send(topic, key=key, value=dict_)

    def write_addition(self, object_type: str, object_: ModelObject) -> None:
        """Write a single object to the journal"""
        self._write_addition(object_type, object_)
        self.flush()

    write_update = write_addition

    def write_additions(self, object_type: str, objects: Iterable[ModelObject]) -> None:
        """Write a set of objects to the journal"""
        for object_ in objects:
            self._write_addition(object_type, object_)

        self.flush()



from swh.journal.client import JournalClient
from swh.journal.writer.kafka import KafkaJournalWriter
from swh.journal.serializers import kafka_to_value, key_to_kafka, value_to_kafka


class FakeKafkaMessage:
    def __init__(self, topic, key, value):
        self._topic = topic
        self._key = key_to_kafka(key)
        self._value = value_to_kafka(value)

    def topic(self):
        return self._topic

    def value(self):
        return self._value

    def key(self):
        return self._key

    def error(self):
        return None


class MockedKafkaWriter(KafkaJournalWriter):
    def __init__(self, queue, anonymize: bool = False):
        self._prefix = "prefix"
        self.queue = queue
        self.anonymize = anonymize

    def send(self, topic, key, value):
        msg = FakeKafkaMessage(topic=topic, key=key, value=value)
        self.queue.append(msg)

    def flush(self):
        pass


class MockedKafkaConsumer:
    """Mimic the confluent_kafka.Consumer API, producing the messages stored
       in `queue`.

       You're only allowed to subscribe to topics in which the queue has
       messages.
    """

    def __init__(self, queue):
        self.queue = queue
        self.committed = False

    def consume(self, num_messages, timeout=None):
        L = self.queue[0:num_messages]
        self.queue[0:num_messages] = []
        return L

    def commit(self):
        if self.queue == []:
            self.committed = True

    def list_topics(self, timeout=None):
        return set(message.topic() for message in self.queue)

    def subscribe(self, topics):
        unknown_topics = set(topics) - self.list_topics()
        if unknown_topics:
            raise ValueError("Unknown topics %s" % ", ".join(unknown_topics))

    def close(self):
        pass


class MockedJournalClient(JournalClient):
    def __init__(self, queue, object_types=None):
        self._object_types = object_types
        self.consumer = MockedKafkaConsumer(queue)
        self.process_timeout = None
        self.stop_after_objects = None
        self.value_deserializer = kafka_to_value
        self.stop_on_eof = False
        self.batch_size = 200



# Copyright (C) 2016-2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, Union, overload

import msgpack

from swh.core.api.serializers import msgpack_dumps, msgpack_loads
from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.model import (
    Content,
    Directory,
    Origin,
    OriginVisit,
    Release,
    Revision,
    SkippedContent,
    Snapshot,
)

ModelObject = Union[
    Content, Directory, Origin, OriginVisit, Release, Revision, SkippedContent, Snapshot
]

KeyType = Union[Dict[str, str], Dict[str, bytes], bytes]


# these @overload'ed versions of the object_key method aim at helping mypy figuring
# the correct type-ing.
@overload
def object_key(
    object_type: str, object_: Union[Content, Directory, Revision, Release, Snapshot]
) -> bytes:
    ...


@overload
def object_key(
    object_type: str, object_: Union[Origin, SkippedContent]
) -> Dict[str, bytes]:
    ...


@overload
def object_key(object_type: str, object_: OriginVisit) -> Dict[str, str]:
    ...


def object_key(object_type: str, object_) -> KeyType:
    if object_type in ("revision", "release", "directory", "snapshot"):
        return object_.id
    elif object_type == "content":
        return object_.sha1  # TODO: use a dict of hashes
    elif object_type == "skipped_content":
        return {hash: getattr(object_, hash) for hash in DEFAULT_ALGORITHMS}
    elif object_type == "origin":
        return {"url": object_.url}
    elif object_type == "origin_visit":
        return {
            "origin": object_.origin,
            "date": str(object_.date),
        }
    elif object_type == "origin_visit_status":
        return {
            "origin": object_.origin,
            "visit": str(object_.visit),
            "date": str(object_.date),
        }
    else:
        raise ValueError("Unknown object type: %s." % object_type)


def stringify_key_item(k: str, v: Union[str, bytes]) -> str:
    """Turn the item of a dict key into a string"""
    if isinstance(v, str):
        return v
    if k == "url":
        return v.decode("utf-8")
    return v.hex()


def pprint_key(key: KeyType) -> str:
    """Pretty-print a kafka key"""

    if isinstance(key, dict):
        return "{%s}" % ", ".join(
            f"{k}: {stringify_key_item(k, v)}" for k, v in key.items()
        )
    elif isinstance(key, bytes):
        return key.hex()
    else:
        return key


def key_to_kafka(key: KeyType) -> bytes:
    """Serialize a key, possibly a dict, in a predictable way"""
    p = msgpack.Packer(use_bin_type=True)
    if isinstance(key, dict):
        return p.pack_map_pairs(sorted(key.items()))
    else:
        return p.pack(key)


def kafka_to_key(kafka_key: bytes) -> KeyType:
    """Deserialize a key"""
    return msgpack.loads(kafka_key, raw=False)


def value_to_kafka(value: Any) -> bytes:
    """Serialize some data for storage in kafka"""
    return msgpack_dumps(value)


def kafka_to_value(kafka_value: bytes) -> Any:
    """Deserialize some data stored in kafka"""
    value = msgpack_loads(kafka_value)
    if isinstance(value, list):
        return tuple(value)
    if isinstance(value, dict):
        return ensure_tuples(value)
    return value


def ensure_tuples(value: Dict) -> Dict:
    return {k: tuple(v) if isinstance(v, list) else v for k, v in value.items()}
